# encoding=CP1251
#from xml.etree.ElementTree import dummy
import nltk, os, math
import cPickle

numberUnigrams = 3
numberBigrams = 4
numberThreegrams = 1
dataPath = "./open/data/"
pathToCollection = "./open/"
pathToAnswers = "./open/answers/"
pathToModel = "./models/ngram.pkl"


# all methods for text processing
class TextProcessor():
    def __init__(self, text):
       # lit = text.rfind(unicode('', "cp1251"))
      #  if lit > 0 :
       #     text = text[:lit]
        allWords = nltk.PunktWordTokenizer().tokenize(text)
        words = filter(lambda x: len(x) > 3, allWords) # filter 1-2 symbols words
        words = words[:230]#+  words[-230:]
        stemmer = nltk.stem.snowball.RussianStemmer()
        self.words = map(stemmer.stem, words)  # apply stemmer to each word
        self.originalForm = self.saveOriginalFormOfWords(words)
    
    # save original form of word
    # here it's first occurred word of the text corresponded to the particular token
    def saveOriginalFormOfWords(self, tokenizedText):            
        originalForm = dict()
        for i in xrange(len(self.words)):
            if not(originalForm.has_key(self.words[i])):
                originalForm[self.words[i]] = tokenizedText[i]
        return originalForm  
    
    # simple word restorer
    def wordRestorer(self, word):
        return self.originalForm[word]
        
# training of algorithms
# here we only compute IDf
class Trainer():
    # model is a pair of (dict: word -> idf_score) and (number of files in training set)
    model = None

    def __init__ (self):
        if not(self.loadModel()):
            self.model = self.computeIDF()
            self.saveModel()
        
    def computeIDF(self):
        fileCounter = 0
        docs = dict() 
        for fileName in os.listdir(pathToCollection):
            print fileCounter,  # progress
            fileCounter += 1
                        
            f = open(pathToCollection + fileName)
            text = unicode(f.read(), "utf-8")
            f.close()
            
            # preprocess text
            tp = TextProcessor(text) 
            
            # process unigrams
            for unigram in set(tp.words):
                docs[unigram] = docs.setdefault(unigram, 0.0) + 1.0
            
            # process bigrams
            for bigram in set(NgramProcessor().getBigrams(tp.words)):
                docs[bigram] = docs.setdefault(bigram, 0.0) + 1.0

            # process threegrams
            for threegram in set(NgramProcessor().getThreegrams(tp.words)):
                docs[threegram] = docs.setdefault(threegram, 0.0) + 1.0


                
            
        # compute IDF
        idf = dict()
        for ngram in docs.keys():
            if docs[ngram] > 1:
                idf[ngram] = -math.log(docs[ngram] / fileCounter)
            
        return (idf, fileCounter)
    
    def saveModel(self):
        f = open(pathToModel, "wb")
        cPickle.dump(self.model, f)
        f.close()
    
    def loadModel(self):
        try:
            f = open(pathToModel)
            self.model = cPickle.load(f)
            f.close()
            return True
        except:
            return False
        
    def getModel(self):
        return self.model


# training of algorithms
# here we only compute IDf
class Trainer2():
    # model is a pair of (dict: word -> idf_score) and (number of files in training set)
    model = None

    def __init__ (self, files):
        self.model = self.computeIDF(files)
        self.saveModel()

    def computeIDF(self, files):
        fileCounter = 0
        docs = dict()
        for fileName in files:
            print fileCounter,  # progress
            fileCounter += 1

            f = open(dataPath + fileName)
            text = unicode(f.read(), "utf-8")
            f.close()

            # preprocess text
            tp = TextProcessor(text)

            # process unigrams
            for unigram in set(tp.words):
                docs[unigram] = docs.setdefault(unigram, 0.0) + 1.0

            # process bigrams
            for bigram in set(NgramProcessor().getBigrams(tp.words)):
                docs[bigram] = docs.setdefault(bigram, 0.0) + 1.0

            # process threegrams
            for threegram in set(NgramProcessor().getThreegrams(tp.words)):
                docs[threegram] = docs.setdefault(threegram, 0.0) + 1.0


        # compute IDF
        idf = dict()
        for ngram in docs.keys():
            if docs[ngram] > 1:
                idf[ngram] = -math.log(docs[ngram] / fileCounter)

        return (idf, fileCounter)

    def saveModel(self):
        f = open(pathToModel, "wb")
        cPickle.dump(self.model, f)
        f.close()

    def getModel(self):
        return self.model

# all functions for processing n-gramms
class NgramProcessor():
    def getBigrams(self, wordList):
        return zip(wordList[:-1], wordList[1:])

    def getThreegrams(self, wordList):
        return zip(wordList[:-2], wordList[1:-1], wordList[2:])
        
    # return sorted list of pairs (object, number of its occurrences in the list)
    # it is used for counting unigrams and bigrams in text
    def objectCounter(self, listOfObjects):
        return [(obj, listOfObjects.count(obj)) for obj in set(listOfObjects)] 
        
    # return TopN object (unigram of bigram) from list    
    def getTopN(self, listOfObjectsWithCounts, N):
        sortedListOfObjectsWithCounts = sorted(listOfObjectsWithCounts, key=lambda x:-int(x[1]))[:N]
        # return sortedListOfObjectsWithCounts
        return [item[0] for item in sortedListOfObjectsWithCounts]    

    # Assign weights to unigrams and bigrams
    # here we use TF*IDF weighting scheme
    def weightingScheme(self, listOfCountedObjects, model):
        return [(objectName, objectCount * model[0].setdefault(objectName, -math.log(1.0 / model[1]))) 
                for objectName, objectCount in listOfCountedObjects]

class Measures():
    def __init__ (self):
        self.p = 0
        self.r = 0
        self.f1 = 0
        self.count = 0;

    def evalMeasures(self, fileName, keywords):
        stemmer = nltk.stem.snowball.RussianStemmer()
        mywords = keywords
        for i in xrange(len(mywords)) :
            mywords[i] = ' '.join(map(stemmer.stem, mywords[i].split(' ')))

        f = open(pathToAnswers + fileName)
        text = unicode(f.read(), "utf-8")
        f.close()
        words =  text.split('\n')
        for i in xrange(len(words)) :
            words[i] = ' '.join(map(stemmer.stem, words[i].split(' ')))

        tp = float(len(set(words) & set(mywords)))
        dp = tp / len(mywords)
        dr = tp / len(words)
        if tp != 0.0 :
            df1 = 2 * dp * dr / (dp + dr)
        else :
            df1 = 0

        self.count += 1
        self.p += dp
        self.r += dr
        self.f1 += df1

    def getMeasures(self):
        return [self.p / self.count, self.r / self.count, self.f1 / self.count]

class CrossValidation():
    def __init__(self, k):
        self.k = k

    def validate(self):
        files = os.listdir(dataPath);
        trainNum = self.k
        testNum = len(files) - trainNum

        measures = Measures()
        iter = 0
        while testNum * iter  < len(files) :
            testFiles = files[iter * testNum : min((iter + 1) * testNum, len(files))]
            trainFiles = files[: iter * testNum] + files[min((iter + 1) * testNum, len(files)) :]
            trainer = Trainer2(trainFiles)
            iter += 1;
            for fName in testFiles :
                f = open(dataPath + fName)
                text = unicode(f.read(), "utf-8")
                f.close()
                kw = Keywords2(trainer)
                reskw = kw.getKeywords(text)
                measures.evalMeasures(fName, reskw)
                print
                print fName
                print measures.getMeasures()
        print
        print measures.getMeasures()

class Keywords():
    
    def __init__(self):
        # load trained model
        self.model = Trainer().getModel()

    def getKeywords(self, text):
        # text processor
        tp = TextProcessor(text)

        ng = NgramProcessor()
        bigrams = ng.getBigrams(tp.words)
        threegrams = ng.getThreegrams(tp.words)

        unigramKeywords = ng.getTopN(ng.weightingScheme(ng.objectCounter(tp.words), self.model),
                                     numberUnigrams)
        bigramKeywords = ng.getTopN(ng.weightingScheme(ng.objectCounter(bigrams),self.model),
                                    numberBigrams)
        threegramKeywords = ng.getTopN(ng.weightingScheme(ng.objectCounter(threegrams),self.model),
                                       numberThreegrams)

        unigramKeywords = map(tp.wordRestorer, unigramKeywords)
        bigramKeywords = [tp.wordRestorer(word[0]) + " " + tp.wordRestorer(word[1])
                          for word in bigramKeywords]
        threegramKeywords = [tp.wordRestorer(word[0]) + " " + tp.wordRestorer(word[1]) + " " + tp.wordRestorer(word[2])
                             for word in threegramKeywords]

        return unigramKeywords + bigramKeywords + threegramKeywords#, [x for x in uniDel]]

class Keywords2():
    def __init__(self, trainer):
        # load trained model
        self.model = trainer.getModel()


    def getKeywords(self, text):
        # text processor
        tp = TextProcessor(text)

        ng = NgramProcessor()
        bigrams = ng.getBigrams(tp.words)
        threegrams = ng.getThreegrams(tp.words)

        unigramKeywords = ng.getTopN(ng.weightingScheme(ng.objectCounter(tp.words), self.model),
            numberUnigrams)
        bigramKeywords = ng.getTopN(ng.weightingScheme(ng.objectCounter(bigrams),self.model),
            numberBigrams)
        threegramKeywords = ng.getTopN(ng.weightingScheme(ng.objectCounter(threegrams),self.model),
            numberThreegrams)

        """
        uniDel = set()
        for biKey in bigramKeywords :
            if biKey[0] in unigramKeywords and biKey[1] in  unigramKeywords :
                uniDel.add(biKey[0])
                uniDel.add(biKey[1])
        print repr([x for x in uniDel]).decode("unicode_escape").encode("UTF-8")
        unigramKeywords = filter(lambda x : x not in uniDel, unigramKeywords)
        """

        unigramKeywords = map(tp.wordRestorer, unigramKeywords)
        bigramKeywords = [tp.wordRestorer(word[0]) + " " + tp.wordRestorer(word[1])
                          for word in bigramKeywords]
        threegramKeywords = [tp.wordRestorer(word[0]) + " " + tp.wordRestorer(word[1]) + " " + tp.wordRestorer(word[2])
                             for word in threegramKeywords]

        return unigramKeywords + bigramKeywords + threegramKeywords#, [x for x in uniDel]]


#cv =  CrossValidation(135).validate()

#trainer = Trainer();
"""
files = os.listdir(dataPath);
measures = Measures()
for fName in files :
    f = open(dataPath + fName)
    text = unicode(f.read(), "utf-8")
    f.close()
    kw = Keywords2(trainer)
    reskw = kw.getKeywords(text)
    measures.evalMeasures(fName, reskw)
    print
    print fName
    print measures.getMeasures()
print
print measures.getMeasures()
"""
"""
kw = Keywords()
for ansName in ["f_ee5b-arpg6t1gbu5.txt"]:#os.listdir(pathToAnswers):
    fd = open(dataPath + ansName)
    textd = unicode(fd.read(), "utf-8")
    fd.close()
    [my, delUni] = kw.getKeywords(textd)
    f = open(pathToAnswers + ansName)
    text = unicode(f.read(), "utf-8")
    f.close()
    words =  text.split('\n')
   # words = filter(lambda x: len(x.split(' ')) > 1, words) # filter 1-2 symbols words
    stemmer = nltk.stem.snowball.RussianStemmer()
    delUni = filter(stemmer.stem, delUni)
    for i in xrange(len(my)) :
            my[i] = ' '.join(map(stemmer.stem, my[i].split(' ')))
    for i in xrange(len(words)) :
        words[i] = ' '.join(map(stemmer.stem, words[i].split(' ')))

    if len(set(delUni)) > 0 :
        print ansName + " " + str(len(set(delUni) & set(words)) > 0)
        print repr(delUni).decode("unicode_escape").encode("UTF-8")
        print repr(words).decode("unicode_escape").encode("UTF-8")
        print repr(my).decode("unicode_escape").encode("UTF-8")


f = open(pathToAnswers + 'f_eduz-arpg6pb4y91.txt')
text = unicode(f.read(), "utf-8")
f.close()
words =  text.split('\n')
"""
## -=test=-
#trainer = Trainer();
"""
kw = Keywords()
fileName = "f_edut-arpg6papehz.txt"
pathD = "./open/data/"
f = open(pathD + fileName)
text = unicode(f.read(), "utf-8")
f.close()
tp = TextProcessor(text)

keywords = kw.getKeywords(text)

print repr(keywords).decode("unicode_escape").encode("cp1251")
print repr(keywords).decode("unicode_escape").encode("UTF-8")
"""


